////////////////////////////////////////////////////////////////////////////////
#ifndef TFuzzySet_h
#define TFuzzySet_h
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Copyright (c)1998  Ali Adams, Bournemouth University, UK.
////////////////////////////////////////////////////////////////////////////////
//
// FuzzyLogic Class Library
//
// TVariable
//		TFuzzySet	eHedge
//	  			TPremise  eOperator
//						TPredicate	TConsequent
//								TFuzzyRule
//										TFuzzyController
//												TFuzzySystem
//
// 		TVariable		==	variable as base for a fuzzyset	(Temperature, in Celsius)
//		TFuzzySet		==	fuzzy value set of a variable
//		eHedge			==	linguistic modifier for a fuzzy set	(e.g. VERY, SOMEWHAT, ...)
// 		scalar			==	input variable's value or defuzzified output variable's value set
//		TPremise		==  scalar is (member of) fuzzyset (e.g. Temperature is HOT)
//		eOperator		==	premises connector {AND, OR}
//		TPredicate		==	Premise eOperator Premise eOperator Premise ...
//		TConsequent		==	Premise
//		TFuzzyRule		==	Predicate  or  if Predicate then Consequent
//		TFuzzyController==  FuzzyRules for a single output variable
//		TFuzzySystem	==  all FuzzyControllers in the system
//
//
////////////////////////////////////////////////////////////////////////////////
//	Fuzzy Logic Processing Sequence:
//
//	1. Aggregation		of predicate truths for each rule
//	2. Correlation		of predicate truths for each rule to consequent
//	3. Composition		of weighted-consequents of all rules
//	4. Defuzzification	of composition's result into a crisp value
////////////////////////////////////////////////////////////////////////////////
//
//
// HELP YOURSELF BY HELPING OTHERS
//
// ThanQ for using this software, feel free to modify, enhance, extend, and/or
// redistribute it as long as you include this header with no modification(s),
// you may add your own header after all other headers of other people (if any).
//
// No warranty of any kind explicit or implicit is given, this software is
// provided "as is" and is only to be used at your own risk, and the author
// does not except any responsibility or liability to any kind of defects,
// damages, or injuries resulting from the (in)direct use of this software.
//
// However, I worked HARD and LONG to get this software to this level of
// usability, readability, and maintainability. The author appreciates any
// reciprocal help from users of this software. This version of the software is
// released as a shareware and any income resulting from it will help sustain
// the author's ability to produce useful, and usable software shared by all.
////////////////////////////////////////////////////////////////////////////////
//
// Please email bugs/fixes/suggestions to:		aliadams@bournemouth.ac.uk
//
////////////////////////////////////////////////////////////////////////////////
// References:
//   1. The Fuzzy Systems Handbook, by Earl Cox, 1994. ISBN 0-12-194270-8
//	 2. The CFuzzySet classes (C) Timothy C. Lee 1996
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
#include<iostream.h>			//	istream, ostream
#include<string.h>				//  strlen(), strcpy()
#include<vcl\graphics.hpp>		//	TColor
#include<vcl\extctrls.hpp>		//	TPaintBox


#include<math.h>				//  pow, sqrt, ...
#include<values.h>				//  MAXDOUBLE, ...
#define		TOLERANCE			0.0001
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// MACROS
////////////////////////////////////////////////////////////////////////////////
#define		min(x,y)		((x)<(y))?(x):(y)
#define		max(x,y)		((x)>(y))?(x):(y)
////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////
class TFuzzySet;	//forward decleration
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//
#define	MIN_TRUTH					0.0		// minimum fuzzy value
#define	MAX_TRUTH					1.0		// maximum fuzzy value
#define	DOMAIN_WIDTH				100		// number of divisions in domain
#define	MAX_FUZZYSETS_PER_VARIABLE	9		// max number of fuzzy values for variable

////////////////////////////////////////////////////////////////////////////////
//	TVariable	==	variable as base for a fuzzyset	(Temperature, in Celsius)
////////////////////////////////////////////////////////////////////////////////
//
class TVariable
{
public:
	char*		name;   // variable name	e.g.	Temperature
	char*		unit;   // measure unit		e.g.	Celsius
	double		begin;	// domain begin     e.g.	-10
	double		end;	// domain end		e.g.	+40

    TFuzzySet*	fuzzysets[MAX_FUZZYSETS_PER_VARIABLE];

public:
	TVariable(char* _name=NULL, char* _unit=NULL, double _begin=0.0*DOMAIN_WIDTH, double _end=1.0*DOMAIN_WIDTH);
	virtual ~TVariable();

	void set_name(char*  _name)	{ if (name != NULL) delete[] name; name = new char[strlen(_name)+1]; strcpy(name, _name); }
	void set_unit(char*  _unit)	{ if (unit != NULL) delete[] unit; unit = new char[strlen(_unit)+1]; strcpy(unit, _unit); }

	void clear_fuzzysets();
	void add_fuzzyset(TFuzzySet* fuzzyset);
	void remove_fuzzyset(TFuzzySet* fuzzyset);

    bool operator == (TVariable& variable2);
};
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
//
#define	DOMAIN_POINTS	DOMAIN_WIDTH + 1		// number of domain points for DOMAIN_WIDTH
#define INVALID_DOMAIN	-MAXDOUBLE				// for all points outside min .. max
#define ZERO_TRUTH		MIN_TRUTH						// for all points outside min .. max
//
//   A fuzzy set's domain lies inside its variable's domain (begin..end), and is
// 	 divided into 4 regions using five points (min, mid1, mid, mid2, max).
//
//	     |......|-------------|--------|-----|------------|.............|
//     begin   min           mid1     mid   mid2        max            end
//
//   In order to keep to one general class for all shapes, these five points are
//   required by constructor, even if some aren't used. You can howvever, pass
//   only the first neccssory few arguments, and let the others take the default
//   valus specified in the constructor.
//
//	 The variable's domain will be sampled using DOMAIN_POINTS points.
//
//   The truth value of the DOMAIN_POINTS ponits will be interpolated using
//		1. the above five points,
//		2. the minTruth & maxTruth values, and
//	 	3. the shape of the fuzzy set.
//	 This will represent the membership function for the fuzzy set.
//
////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////
// truth function's shape directions
////////////////////////////////////////////////////////////////////////////////
#define UP					0
#define DOWN				1
#define LEFT 				DOWN		// for SHOULDER shapes
#define RIGHT				UP			// for SHOULDER shapes
////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// truth function's shapes
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//enum	eHedge	{NO_SHAPE, SINGLETON, PLATEAU, LINEAR, TRIANGLE, TRAPEZOIDAL, SHOULDER, S_CURVE, PI_CURVE, BETA_CURVE, GAUSSIAN_CURVE, IRREGULAR};
//
#define NO_SHAPE			0	// FuzzySet          ()
//	parameters ignored			   					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// truth(x)  = MIN_TRUTH
//  typically, create empty set with NO_SHAPE as an output set, use it to store the result of fuzzyset operations
#define SINGLETON			1	// FuzzySinglton     (           mid,            type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	parameters ignored			   					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// for  mid  >  x			--> truth(x) = minTruth
//                |                                                                                                    	// for  mid  == x+-TOLERANCE--> truth(x) = maxTruth
//                |                                                                                                  	// for  mid  >  x			--> truth(x) = minTruth
//     -----------|----------
#define PLATEAU				2	// FuzzyPlateau      (min,                  max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
#define BOOLEAN				PLATEAU
//	mid1,mid,mid2,type parameters ignored			 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// for  min  >  x			--> truth(x) = minTruth
//     |---------------------|                                                                                    		//      min  <= x <= max	--> truth(x) = maxTruth
//	   |                     |                                                                                    		//              x >  max	--> truth(x) = minTruth
//     |                     |
#define LINEAR				3	// FuzzyLine         (min,                  max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	mid1,mid,mid2 parameters ignored				 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// for  min  >  x	  		--> truth(x) = minTruth
//		                _    *      																			  		//	  	min  <= x <= max 	--> truth(x) = x * ((maxTruth-minTruth)/(max-min))
//          _     *            UP                                                                                   	//              x >  max 	--> truth(x) = maxTruth
//     *
//	mid1,mid,mid2 parameters ignored				 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// 1 - LINE_UP
//	   *    _                     																			  			//
//                *     _       DOWN                                                                                   	//
//                           *
#define TRIANGLE			4	// FuzzyTriangle     (min,       mid,       max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	mid1,mid2 parameters ignored 					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// for  min  >  x			--> truth(x) = minTruth
//			      *  																							  		//	  	min  <= x <= mid 	--> truth(x) = x * ((maxTruth-minTruth)/(mid-min))
//			    *   *                                                                                             		//	  	mid  <= x <= max 	--> truth(x) = MAX_TRUTH - x * ((maxTruth-minTruth)/(max-mid))
//           *         *                                                                                          		//              x >  max 	--> truth(x) = minTruth
//        *               *
//     *                     *
#define TRAPEZOIDAL			5	// FuzzyTrapezodal   (min, mid1,      mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	mid parameters ignored		   					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH); 	// for  min  >  x		  	--> truth(x) = minTruth
//			 *---------*																						  		//	  	min  <= x <= mid1 	--> truth(x) = x * ((maxTruth-minTruth)/(mid1-min))
//		   *			 *																						  		//	  	mid1 <= x <= mid2 	--> truth(x) = maxTruth
//       *                 *                                                                                      		//	  	mid2 <= x <= max  	--> truth(x) = MAX_TRUTH - x * ((maxTruth-minTruth)/(max-mid2))
//     *                     *                                                                                    		//              x >  max  	--> truth(x) = minTruth
#define SHOULDER			6	// FuzzyShoulder     (min,     , mid,     , max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	mid1,mid2 parameters ignored					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);  // narrow(LINE_UP, 0.25) + shift_right(narrow(PLATEAU, 0.25), 0.25)
//                      *-----
//   type=UP=RIGHT    *
//                  *
//     -----*
//            *    type=DOWN=LEFT
//              *
#define S_CURVE				7	// FuzzySCurve       (min,       flex,      max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	mid1,mid2 parameters ignored, flex=mid			 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);	// for          x <= min    -->  truth(x) = minTruth
//				 	    .  ` `																					  		//       min <= x <= mid	-->  truth(x) = minTruth + 2*(((x-min)/((mid-min)*2))^2)
//		type=UP      *																									//       mid <= x <= max	-->  truth(x) = maxTruth + 2*(((max-x)/((max-mid)*2))^2)
//		          *																									  	//              x >= max	-->  truth(x) = maxTruth
//            *
//     . . `
//	mid1,mid1 parameters ignored, flex=mid			 (min, mid1, mid, mid2, max, type=DOWN, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);// 1 - S_CURVE_UP
//	   ` `	.
//			   *
//		type=DOWN *
//                   *
//                      `  . .
#define PI_CURVE			8	// FuzzyPICurve      (min,       mid,       max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	type parameter ingored		   					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);  // narrow(S_CURVE_UP, 0.5) + shift_right(narrow(S_CURVE_DOWN, 0.5), 0.5)
//				. `` .																									//
//			  *        *                                                                                                //
//		     *          *                                                                                               //
//          *            *                                                                                              //
//      . `                ` .
#define BETA_CURVE			9	// FuzzyBetaCurve    (min, flex1,mid, flex2,max,  type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	min,max,type parameters ingored		 			 (min, mid1, mid, mid2, max,  type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);	// for   min >= x           -->  approaching -infinity
//				. `` .                                                                                                  //       min <= x <= mid	-->  truth(x) = 1 / ( 1 + ((mid-x)/mid1)^2 )
//			  *        *                                                                                                //       mid <= x <= max	-->  truth(x) = 1 / ( 1 + ((x-mid)/mid2)^2 )
//		     *          *                                                                                               //              x >= max    -->  approaching +infinity
//          *            *
//      ` `                ` `
#define GAUSSIAN_CURVE		10	// FuzzyGaussianCurve(centre, width,        type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH)
//	centre=mid, width=mid-min	   					 (min, mid1, mid, mid2, max, type=UP, minTruth=MIN_TRUTH, maxTruth=MAX_TRUTH);  // for          x <= min    -->  truth(x) = minTruth
//				. `` .                                                                                                  //       min <= x <= mid	-->  truth(x) = e ^ ( -(mid-min) * (mid-x)^2 )
//			  *        *                                                                                                //       mid <= x <= max	-->  truth(x) = e ^ ( -(max-mid) * (x-mid)^2 )
//		     *          *                                                                                               //              x >= max    -->  truth(x) = minTruth
//          *            *
//        .`              `.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// linguistic hedges
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define		MAX_HEDGES				16	// maximum number of hedges allowed per set
#define 	INVALID_HEDGE			-1	// for all empty elements in hedges[]



// negation
#define		NOT						2	// !	// as defined in eOperator in FuzzyRules.h

// fuzzify a scalar, or widen
#define		ABOUT					3	// fuzzify(singlton), or widen
#define		AROUND					ABOUT
#define		ROUGHLY					ABOUT

// restrict
#define		ABOVE					4	// shift_right(50%)
#define		LESS_THAN	 			ABOVE

// restrict
#define		BELOW					5	// shift_left(50%)
#define		MORE_THAN				BELOW

// contrast intensification (rotate_anti_clock_wise)
#define		POSITIVELY				6	// if truths[x]<=0.5 --> n*((  x)^n), if x>=mid --> 1 - n*((1-x)^n) , [n=2]
#define		ALMOST					POSITIVELY
#define		DEFINITELY				POSITIVELY

// contrast diffusion  (rotate_clock_wise)
#define		GENERALLY				7	// if truths[x]> 0.5 --> ((1/n)*(  x))^(1/n), if x>=mid --> 1 - ((1/n)*(1-x))^(1/n) , [n=2]
#define		USUALLY					GENERALLY

// approximate broadly
#define		IN_THE_VICINITY_OF		8	// widen a lot

// approximate narrowly
#define		CLOSE_TO				9	// narrow a lot
#define		NEIGHBORING				CLOSE_TO

// dilute
#define		SLIGHTLY				10	// x^(1/n)	, [n=3]
#define		SOMEWHAT				11	// x^(1/n)	, [n=2]
#define		RATHER					SOMEWHAT
#define		QUITE					SOMEWHAT

// intensify ^n
#define		VERY					12	// x^n		, [n=2]
#define		EXTREMELY				13	// x^n		, [n=3]
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// alpha cut types
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define STRONG				0	// restrict domain to x:truth(x) >   alphaCut
#define WEAK				1	// restrict domain to x:truth(x) >=  alphaCut
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// logical operators implementation methods for			and()							or()							not()
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define ZADEH				0	//					min(s1,s2)						max(s1,s2)							1-s1
#define MEAN				1	//      	    	(s1+s2)/2				(2*min(s1,s2)+4*max(s1,s2))/6               1-s1
#define MEANSQUARE			2	//         			((s1+s2)/2)^2			((2*min(s1,s2)+4*max(s1,s2))/6)^2           1-s1
#define MEANROOT			3	//					((s1+s2)/2)^.5			((2*min(s1,s2)+4*max(s1,s2))/6)^.5          1-s1
#define PRODUCT				4	//						s1*s2						(s1+s2)-(s1*s2)                     1-s1
#define BOUNDEDSUM			5	//					max(0,(s1+s2-1))				min(1,(s1+s2))
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// aggregation methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// use logical operators for ANDing
#define ZADEH				0
#define MEAN				1
#define MEANSQUARE			2
#define MEANROOT			3
#define PRODUCT				4
#define BOUNDEDSUM			5
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// correlation methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define TRUNCATE 				0	// limit consequence's maxTruth to predicates-aggregated value
#define SCALE					1	// scale-down consequence by the predicates-aggregated value
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// composition methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// use logical operators for ORing
#define ZADEH				0
#define MEAN				1
#define MEANSQUARE			2
#define MEANROOT			3
#define PRODUCT				4
#define BOUNDEDSUM			5
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// defuzzification methods
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#define CENTROID				0					//find domain point x at the centre of gravity ({Xi*Yi}/{Yi})
#define COMPOSITE_MOMENTS		CENTROID
#define COMPOSITE_MAXIMUM		1					//find domain point x with maximum truth, if plateau, then use either AVERAGE_MAXIMUM, or CENTRE_OF_MAXIMUMS
#define MAXIMUM_HEIGHT			COMPOSITE_MAXIMUM
#define NEAR_EDGE				2					//find x at the left  edge of usually a single-edged plateau
#define FAR_EDGE				3					//find x at the right edge of usually a single-edged plateau
#define CENTROID_SUPPORT_SET	4 					//similar to CENTROID, but for only the part of the domain with truth(X) > alphaCut (STRONG)
#define AVERAGE_MAXIMUM			5 					//similar to MAXIMUM_HEIGHT, but for only the highest plateau, or singlton
#define CENTRE_OF_MAXIMUMS		6	 				//similar to MAXIMUM_HEIGHT, but for only the two highest plateaus, or singltons
#define CENTRE_OF_MASS			7 					//find x at the centre of the region supported by the maximum number of rules or evidence
#define BEST_EVIDENCE			CENTRE_OF_MASS
#define LEAST_ENTROPY			8					//find x for the least degree of fuzzyness (information entropy)
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////




////////////////////////////////////////////////////////////////////////////////
// TFuzzySet decleration
////////////////////////////////////////////////////////////////////////////////
class TFuzzySet
{
  private:
	TVariable*		variable;
	char* 		 	name;
	int	  		 	shape;
	int	  		 	direction;
	double		 	min;		// start of domain
	double		 	mid1;		// somewhere between min and mid
	double		 	mid;		// near the middle of the domain
	double		 	mid2;		// somewhere between mid and max
	double		 	max;		// end of domain
	double		 	minTruth;
	double		 	maxTruth;
	double			truths[DOMAIN_POINTS];	// membership function points
	int				hedges[MAX_HEDGES];		// hedges applied to fuzzy set
	double		 	alphaCut;               // restrict domain to x:truth(x) >= alphaCut (STRONG defualt)

	void	 		interpolate_truths();

  public:

	TFuzzySet(  TVariable* _variable	= NULL,
    			char* 	   _name		= NULL,
                int 	   _shape		= NO_SHAPE,
                int 	   _direction	= UP,
    			double 	   _min			= 0.00 * DOMAIN_WIDTH,
                double 	   _mid1		= 0.25 * DOMAIN_WIDTH,
                double 	   _mid			= 0.50 * DOMAIN_WIDTH,
                double 	   _mid2		= 0.75 * DOMAIN_WIDTH,
                double 	   _max			= 1.00 * DOMAIN_WIDTH,
                double 	   _minTruth	= MIN_TRUTH,
                double 	   _maxTruth	= MAX_TRUTH
    		 );
	TFuzzySet(TFuzzySet& set2);	// copy constructor (for deep copy)
	TFuzzySet(double scalar);	// convert scalar into a singleton
	virtual ~TFuzzySet();

	TVariable*	get_variable	()			{return variable;}
	char*		get_name		()			{return name;}
	int			get_shape		()			{return shape;}
	int			get_direction	()			{return direction;}
	double		get_min			()			{return ( (min * (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin );}
	double		get_mid1		()			{return ( (mid1* (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin );}
	double		get_mid			()			{return ( (mid * (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin );}
	double		get_mid2		()			{return ( (mid2* (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin );}
	double		get_max			()			{return ( (max * (variable->end - variable->begin))/DOMAIN_WIDTH + variable->begin );}

	double		get_minTruth	()			{return minTruth;}
	double		get_maxTruth	()			{return maxTruth;}
	double		truth         	(double x)	{ return truths[(int)x];}
    int			get_hedge 		(int i)		{return hedges[i];}
	double		get_alphaCut	()			{return alphaCut;}

	void set_variable 	(TVariable* _variable)	{ variable	= _variable; }
	void set_name	 	(char*  _name)          { if (name != NULL) delete[] name; name = new char[strlen(_name)+1]; strcpy(name, _name); }
	void set_shape	 	(int    _shape)			{ shape		= _shape; }
	void set_direction	(int    _direction)		{ direction	= _direction; }
	void set_min 		(double _min )			{ min   	= ((_min  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH ; }
	void set_mid1		(double _mid1)			{ mid1  	= ((_mid1 - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH ; }
	void set_mid 		(double _mid )			{ mid   	= ((_mid  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH ; }
	void set_mid2		(double _mid2)			{ mid2  	= ((_mid2 - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH ; }
	void set_max 		(double _max )			{ max   	= ((_max  - variable->begin)/(variable->end - variable->begin))*DOMAIN_WIDTH ; }
	void set_minTruth	(double _minTruth)		{ minTruth  = _minTruth; }
	void set_maxTruth	(double _maxTruth)		{ maxTruth  = _maxTruth; }
    void clear_hedges	();
    void add_hedge		(int type);
	void remove_hedge	(int type);
    void set_alphaCut	(double _alphaCut)		{ alphaCut	= _alphaCut; }

    // manipulator functions
	void		normalize     	();						// multiply all by MAX_TRUTH/maxTruth() so that maxTruth()=MAX_TRUTH
	bool		normalized    	();                		// is normalised?
	void		truncate      	(double _maxTruth);		// for all truth(x) > _maxTruth make truth(x) = _maxTruth
    void 		apply_alphaCut	(int type=STRONG);
	double		defuzzify	  	(int how=CENTROID);		// convert to a scalar

    // move and resize functions
	void		shift_left		(double percent);
	void		shift_right	  	(double percent);
	void		shift_up	  	(double factor);
	void		shift_down	  	(double percent);
	void		narrow 			(double percent);
	void		widen  			(double factor);
	void		scale_up  		(double factor);
	void		scale_down		(double percent);

    // logical operators (for all Zahed and non_Zadeh styles)
	TFuzzySet&	and			  	(TFuzzySet& set2, int type=ZADEH);
	TFuzzySet&	or 			  	(TFuzzySet& set2, int type=ZADEH);
	TFuzzySet&	not			  	(int type=ZADEH);

			bool       operator == 		(TFuzzySet& set2);   				// comparison
			bool       operator != 		(TFuzzySet& set2);   				// comparison
			bool       operator <  		(TFuzzySet& set2);   				// comparison
			bool       operator >  		(TFuzzySet& set2);   				// comparison
			bool       operator <= 		(TFuzzySet& set2);   				// comparison
			bool       operator >= 		(TFuzzySet& set2);   				// comparison

			TFuzzySet& operator = 		(TFuzzySet& set2);   				// assignment

	friend	TFuzzySet& operator +  		(TFuzzySet& set1, TFuzzySet& set2);	// WARNING: maxTruth can be > MAX_TRUTH
	friend	TFuzzySet& operator +  		(TFuzzySet& set1, double shift);	// shift-up
	friend	TFuzzySet& operator +  		(double shift, TFuzzySet& set1);	// shift-up

	friend	TFuzzySet& operator -  		(TFuzzySet& set1, TFuzzySet& set2);	// WARNING: maxTruth can be > MAX_TRUTH
	friend	TFuzzySet& operator -  		(TFuzzySet& set1, double shift);	// shift-down
	friend	TFuzzySet& operator -  		(double shift, TFuzzySet& set1);	// shift-down

	friend	TFuzzySet& operator *  		(TFuzzySet& set1, TFuzzySet& set2);	// WARNING: maxTruth can be > MAX_TRUTH
	friend	TFuzzySet& operator *  		(TFuzzySet& set1, double factor);	// scale up
	friend	TFuzzySet& operator *  		(double factor, TFuzzySet& set1);	// scale up

	friend	TFuzzySet& operator /  		(TFuzzySet& set1, TFuzzySet& set2);	// WARNING: maxTruth can be > MAX_TRUTH
	friend	TFuzzySet& operator /  		(TFuzzySet& set1, double percent);	// scale down
	friend	TFuzzySet& operator /  		(double percent, TFuzzySet& set1);	// scale down

	friend	TFuzzySet& operator ^  		(TFuzzySet& set1, double power);	// hedge (dilute or intensify)

	friend	TFuzzySet& operator && 		(TFuzzySet& set1, TFuzzySet& set2);	// Zadeh and
	friend	TFuzzySet& operator || 		(TFuzzySet& set1, TFuzzySet& set2);	// Zadeh or
	friend	TFuzzySet& operator !  		(TFuzzySet& set1);					// Zadeh not

			double	   operator [] 		(double x);							// truth(double)
					   operator double 	();  								// defuzzify (convert to double)

	friend	ostream&   operator <<		(ostream& s, TFuzzySet& set1);		// save
	friend	istream&   operator >>		(istream& s, TFuzzySet& set1);		// load


	// graphics
	void   	Draw(TPaintBox* PaintBox);
};
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
#endif
////////////////////////////////////////////////////////////////////////////////

